import urllib.request
import urllib.parse
import json as jsonlib
import time
import asyncio
import base64
from typing import Any, Optional, Dict, Union, AsyncGenerator, Tuple, cast, Iterator
from collections.abc import Iterable
import http.cookiejar
from urllib.error import URLError, HTTPError


class HTTPResponse:
    _raw: Any
    status: int
    headers: Dict[str, str]
    _stream: bool
    _cached_content: Optional[bytes]

    def __init__(self, raw_resp: Any, status: int, headers: Dict[str, str], stream: bool = False) -> None:
        self._raw = raw_resp
        self.status = status
        self.headers = headers
        self._stream = stream
        self._cached_content = None

    def text(self) -> str:
        """Return response body decoded to text respecting charset if any."""
        encoding = "utf-8"
        content_type = self.headers.get("Content-Type", "")
        if "charset=" in content_type:
            try:
                enc = content_type.split("charset=")[1].split(";")[0].strip()
                if enc:
                    encoding = enc
            except Exception:
                pass
        return self.content().decode(encoding, errors="replace")

    def json(self) -> Any:
        """Parse response body as JSON."""
        return jsonlib.loads(self.text())

    def content(self) -> bytes:
        """Return the entire response body as bytes."""
        if self._cached_content is None:
            if self._stream:
                # Read all at once if stream=True by consuming the iterator
                self._cached_content = b"".join(self.iter_content())
            else:
                self._cached_content = self._raw.read()
        return cast(bytes, self._cached_content)

    def iter_content(self, chunk_size: Optional[int] = 1024) -> Iterator[bytes]:
        if self._raw is None:
            yield b""
            return
        
        if not self._stream:
            yield self.content()
            return

        if chunk_size is None:
            # Read all content at once and yield it once
            chunk = self._raw.read()
            if chunk:
                yield chunk
            return

        if chunk_size <= 0:
            # Yield empty bytes and stop
            yield b""
            return

        while True:
            chunk = self._raw.read(chunk_size)
            if not chunk:
                break
            yield chunk

    async def aiter_content(self, chunk_size: Optional[int] = 8192) -> AsyncGenerator[bytes, None]:
        """Async chunked iterator (reads in thread to avoid blocking event loop)."""
        # Defensive fallback if chunk_size is None or invalid
        if chunk_size is None or chunk_size <= 0:
            chunk_size = 8192

        while True:
            chunk = await asyncio.to_thread(self._raw.read, chunk_size)
            if not chunk:
                break
            yield chunk

    def close(self) -> None:
        """Close the underlying raw response if it supports close."""
        close_method = getattr(self._raw, "close", None)
        if callable(close_method):
            close_method()


class Session:
    headers: Dict[str, str]
    params: Dict[str, str]
    timeout: int
    retries: int
    stream: bool
    _cookie_jar: http.cookiejar.CookieJar
    _opener: urllib.request.OpenerDirector
    opener: urllib.request.OpenerDirector

    def __init__(
        self,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        timeout: int = 10,
        retries: int = 3,
        stream: bool = False,
    ) -> None:
        self.headers = dict(headers or {})
        self.params = dict(params or {})
        self.timeout = timeout
        self.retries = retries
        self.stream = stream

        # Cookie handling
        self._cookie_jar = http.cookiejar.CookieJar()
        self._opener = urllib.request.build_opener(
            urllib.request.HTTPCookieProcessor(self._cookie_jar),
            urllib.request.HTTPRedirectHandler(),
        )

        # Allow user to set a custom opener later if desired
        self.opener = self._opener

    def _apply_auth_header(self, headers: Dict[str, str], auth: Optional[Tuple[str, str]]) -> None:
        if auth:
            user, pwd = auth
            token = base64.b64encode(f"{user}:{pwd}".encode("utf-8")).decode("ascii")
            headers.setdefault("Authorization", f"Basic {token}")

    def request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, str]] = None,
        data: Optional[Union[dict, str, bytes]] = None,
        json: Optional[Any] = None,  # <-- added json parameter
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        retries: Optional[int] = None,
        stream: Optional[bool] = None,
        auth: Optional[Tuple[str, str]] = None,
    ) -> HTTPResponse:
        # Merge defaults
        timeout = self.timeout if timeout is None else timeout
        retries = self.retries if retries is None else retries
        stream = self.stream if stream is None else stream

        # Build headers and params
        final_headers = dict(self.headers)  # session headers
        if headers:
            final_headers.update(headers)
        self._apply_auth_header(final_headers, auth)

        final_params = dict(self.params)
        if params:
            final_params.update(params)
        if final_params:
            url = url + ("&" if "?" in url else "?") + urllib.parse.urlencode(final_params)

        # Prepare body
        body_bytes: Optional[bytes] = None

        if json is not None:
            # JSON mode takes precedence over data
            final_headers.setdefault("Content-Type", "application/json")
            body_bytes = jsonlib.dumps(json).encode("utf-8")
        elif data is not None:
            if isinstance(data, dict):
                # Default: form-encoded
                final_headers.setdefault("Content-Type", "application/x-www-form-urlencoded")
                body_bytes = urllib.parse.urlencode(data).encode("utf-8")
            elif isinstance(data, str):
                body_bytes = data.encode("utf-8")
                final_headers.setdefault("Content-Type", "application/x-www-form-urlencoded")
            elif isinstance(data, (bytes, bytearray, memoryview)):
                # Accept bytearray and memoryview as raw bytes
                body_bytes = bytes(data)  # convert to bytes if needed
            elif hasattr(data, "read") and callable(data.read):
                # Assume file-like, read bytes
                body_bytes = data.read()
                if not isinstance(body_bytes, bytes):
                    raise TypeError("file-like object's read() must return bytes")
            elif isinstance(data, Iterable):
                # Accept iterable of bytes chunks; join them
                body_bytes = b"".join(data)
            else:
                raise TypeError("data must be dict, str, or bytes")

        last_exc: Optional[Exception] = None
        for attempt in range(retries):
            try:
                req = urllib.request.Request(
                    url, data=body_bytes, headers=final_headers, method=method.upper()
                )
                raw_resp = self.opener.open(req, timeout=timeout)
                return HTTPResponse(raw_resp, raw_resp.getcode(), dict(raw_resp.getheaders()), stream=stream)
            except (HTTPError, URLError, OSError) as e:
                last_exc = e
                if isinstance(e, HTTPError):
                    raw_resp = cast(Any, e)
                    return HTTPResponse(raw_resp, e.code, dict(e.headers or {}), stream=stream)
                if attempt < retries - 1:
                    time.sleep(1)
                    continue
                raise last_exc

        # Defensive fallback (should never reach here)
        raise RuntimeError("Request failed unexpectedly without raising an exception")

    # Convenience sync methods
    def get(self, url: str, **kwargs: Any) -> HTTPResponse:
        return self.request("GET", url, **kwargs)

    def post(self, url: str, **kwargs: Any) -> HTTPResponse:
        return self.request("POST", url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> HTTPResponse:
        return self.request("PUT", url, **kwargs)

    def patch(self, url: str, **kwargs: Any) -> HTTPResponse:
        return self.request("PATCH", url, **kwargs)

    def delete(self, url: str, **kwargs: Any) -> HTTPResponse:
        return self.request("DELETE", url, **kwargs)

    # Async wrappers using asyncio.to_thread to avoid blocking the event loop
    async def async_request(self, method: str, url: str, **kwargs: Any) -> HTTPResponse:
        return await asyncio.to_thread(self.request, method, url, **kwargs)

    async def async_get(self, url: str, **kwargs: Any) -> HTTPResponse:
        return await self.async_request("GET", url, **kwargs)

    async def async_post(self, url: str, **kwargs: Any) -> HTTPResponse:
        return await self.async_request("POST", url, **kwargs)

    async def async_put(self, url: str, **kwargs: Any) -> HTTPResponse:
        return await self.async_request("PUT", url, **kwargs)

    async def async_patch(self, url: str, **kwargs: Any) -> HTTPResponse:
        return await self.async_request("PATCH", url, **kwargs)

    async def async_delete(self, url: str, **kwargs: Any) -> HTTPResponse:
        return await self.async_request("DELETE", url, **kwargs)

    # Context manager support
    def __enter__(self) -> "Session":
        return self

    def __exit__(self, exc_type: Optional[type], exc: Optional[BaseException], tb: Optional[Any]) -> bool:
        # nothing special to close; cookiejar/opener don't need explicit close
        return False

    def close(self) -> None:
        """
        Close the session by clearing cookies and closing any underlying resources.
        This is just for API consistency, this isn't needed if using a Context manager"""
        # Clear all cookies
        self._cookie_jar.clear()

        # Try to close the opener if it has a close method (some custom openers might)
        close_method = getattr(self.opener, "close", None)
        if callable(close_method):
            close_method()



from operator import itemgetter
import sys
import json
import logging
from .stdrequests_session import Session
from pathlib import Path
from getpass import getpass
from datetime import datetime, timedelta
from tzlocal import get_localzone
import os
import csv

# Initialize a logger
logger = logging.getLogger(__name__)
logging.basicConfig(
    stream=sys.stdout,
    format="[%(asctime)s] %(levelname)s %(module)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

# TODO: Make log level configurable
logger.setLevel("INFO")

DEFAULT_CONFIG = Path.home() / ".config/woffu/woffu_auth.json"
DEFAULT_DOCS_DIR = Path.home() / "Documents/woffu/docs"
DEFAULT_SUMMARY_REPORTS_DIR = Path.home() / "Documents/woffu/summary_reports"

DEFAULT_DATE_FORMAT = "%Y-%m-%d"

class WoffuAPIClient(Session):
    # Class arguments
    _woffu_api_url: str = "https://app.woffu.com"
    _localzone = get_localzone()
    
    hour_types_dict: dict = {
        5: "Extr. a compensar"
    }

    def _get_domain_user_companyId(self):
        """
        Get the required Company ID, Domain and User ID, required for HTTP requests.
        One-time only call; this data should be stored in a file and reused from there.
        """
        # This function should only be called the first time the script runs.
        # We'll store the results for subsequent executions
        logger.debug("Retrieving Company IDs...")
        
        # First we need the Company ID from the Users information
        users = self.get(
            url=f"{self._woffu_api_url}/api/users"
        ).json()

        # With that, we retrieve the company Domain
        company = self.get(
            url=f"{self._woffu_api_url}/api/companies/{users['CompanyId']}"
        ).json()

        # Set class arguments
        self._domain = company["Domain"]
        self._user_id, self._company_id = itemgetter(
            'UserId',
            'CompanyId'
        )(users)


    def _retrieve_access_token(self, username: str = "", password: str = ""):
        """
        Authentication process to retrieve token
        """
        if not username or not password:
            logger.error("No username or password provided.")
            return
        
        logger.info("Requesting access token...")
        token_response = self.post(
            url=f"{self._woffu_api_url}/token",
            data = {
                "grant_type": "password",
                "username": username,
                "password": password
            }
        )

        if token_response.status == 200:
            self._token = token_response.json()['access_token']
        else:
            logger.error(f"Can't retrieve access token. Please review username and password and try again. Error code: {token_response.status}")
            self._token = ""
        

    def _save_credentials(self):
        """
        Save the credentials in a config file for later use.
        """
        self._config_file.parent.mkdir(parents=True, exist_ok=True)
        self._config_file.write_text(
            data = json.dumps(
                {
                    "username": self._username,
                    "token": self._token,
                    "user_id": self._user_id,
                    "company_id": self._company_id,
                    "domain": self._domain
                }
            )
        )
        logger.info(f"✅ Credentials stored in: {self._config_file}")


    def _request_credentials(self):
        """
        Request all available information to compose credentials
        Use WOFFU_USERNAME and WOFFU_PASSWORD environment variables for unattended request
        """
        self._username = os.environ.get('WOFFU_USERNAME', "")
        password = os.environ.get('WOFFU_PASSWORD', "")

        # Ask for credentials manually if not provided
        if (not self._username or not password):
            if self._interactive:
                self._username = input("Enter your Woffu username (mail):\n")
                password = getpass(prompt="Enter your password:\n")
            else:
                logger.error("Can't request token in non-interactive method without username and password. Please provide them in WOFFU_USERNAME and WOFFU_PASSWORD.")
                sys.exit(1)

        # Retrieve access token
        self._retrieve_access_token(
            username=self._username,
            password=password
        )

        # Set authentication headers
        self.headers=self._compose_auth_headers()
        logger.info("Retrieving Company information...")

        # Get Company information
        self._get_domain_user_companyId()


    def _load_credentials(self, creds_file: str = "") -> None:
        """
        Load Woffu credentials stored in provided file
        """
        # Update the config file path if a new one is provided
        if creds_file:
            self._config_file = Path(creds_file)
        
        if not self._config_file.exists():
            logger.error(f"Config file '{self._config_file}' doesn't exist! Requesting authentication token...")
            self._request_credentials()
            self._save_credentials()
            
        else:
            with open(self._config_file, "r") as f:
                creds_info = json.load(f)
                self._domain, self._username, self._token, self._user_id, self._company_id = itemgetter(
                    "domain",
                    "username",
                    "token",
                    "user_id",
                    "company_id"
                )(creds_info)
                # Set authentication headers
                self.headers=self._compose_auth_headers()


    def _compose_auth_headers(self) -> dict:
        """
        Compose the authentication headers
        """
        return {
            'Authorization': f"Bearer {self._token}",
            'Accept': 'application/json',
        }


    def __init__(self, **kwargs) -> None:
        # Instance arguments
        self._domain: str = ""
        self._username: str = ""
        self._token: str = ""
        self._user_id: str = ""
        self._company_id: str = ""
        self._config_file: Path = Path(kwargs["config"]) if "config" in kwargs else DEFAULT_CONFIG
        self._documents_path : Path = Path(kwargs["documents_path"]) if "documents_path" in kwargs else DEFAULT_DOCS_DIR
        self._interactive: bool = kwargs["interactive"] if "interactive" in kwargs else False

        # Initialize the parent class
        super().__init__()

        # load config file if provided
        self._load_credentials()


    def get_documents(self, page_size: int = 200) -> list[dict]:
        """
        Return a dictionary with the user's available documents
        """

        documents_dict = self.get(
            url=f"https://{self._domain}/api/users/{self._user_id}/all/documents",
            params={
                "visible": "true",
                "pageSize": str(page_size)
            }
        ).json()

        if "Documents" in documents_dict:
            logger.info(f"{documents_dict['TotalRecords']} documents found")
            return documents_dict["Documents"]
        
        logger.warning(f"No documents available for user {self._username}")
        return []
    

    def download_document(self, document: dict, output_dir: str) -> None:
        """
        Download the document to the defined output_path 
        """
        if output_dir:
            output_path: Path = Path(output_dir)
        else:
            output_path: Path = self._documents_path

        # Compose the file path
        #document_path = os.path.join(output_path, document["Name"])
        document_path = Path.joinpath(Path(output_path), document["Name"])

        if document_path.exists():
            logger.debug(f"Document '{document['Name']}' already exists in the documents folder, not downloading again")
            return
        
        # Create output path if it doesn't exist        
        if not output_path.exists():
            logger.debug(f"Creating output directory: {output_path}")
            output_path.mkdir(parents=True, exist_ok=True)

        # Compose the download link
        document_url = f"https://{self._domain}/api/documents/{document['DocumentId']}/download2"

        document_response = self.get(url=document_url)

        # Save the document
        if document_response.status == 200:
            logger.info(f"Saving '{document['Name']}'...")
            #with open(file=document_path, mode='bw') as f:
            #    f.write(document_response.content())
            document_path.write_bytes(document_response.content())

    def download_all_documents(self, output_dir: str = "") -> None:
        """
        Download all user's documents
        """
        # Retrieve the list of available documents
        documents_list = self.get_documents()
        
        # Iterate over all documents and download them
        if documents_list:
            logger.info("Downloading all documents...")
            for document in documents_list:
                self.download_document(document=document, output_dir=output_dir)
            logger.info("All documents downloaded!")

    def _get_presence(self, from_date: str = "", to_date: str = "", page_size: int = 1000) -> dict:
        """
        Return the presence summary of a user within the provided time window. If no dates are defined, they will be initialized to the current date
        params:
        from_date: str. Start of the time window formatted as 'YYYY-mm-dd'
        to_date: str. End of the time window formatted as 'YYYY-mm-dd'
        page_size: int. Number of entries to retrieve. This should match the number of queried days, but we'll leave it at 1000 by default.
        """

        current_date = datetime.now(tz=self._localzone)
        
        # Initialize date values
        if not from_date:
            from_date = current_date.strftime(DEFAULT_DATE_FORMAT)
            to_date = current_date.strftime(DEFAULT_DATE_FORMAT)

        hours_response = self.get(
            url=f"https://{self._domain}/api/svc/core/diariesquery/users/{self._user_id}/diaries/summary/presence",
            params={
                "userId": self._user_id,
                "fromDate": from_date,
                "toDate": to_date,
                "pageSize": page_size,
                "includeHourTypes":	True,
                "includeNotHourTypes": True,
                "includeDifference": True
            }
        )

        if hours_response.status == 200:
            return hours_response.json()['diaries']
    
        else:
            logger.error(f"Can't retrieve presence for the time period {from_date} - {to_date}!")
            return {}


    def _get_diary_hour_types(self, date: str) -> dict:
        """
        Return the hour types' summary for a given date
        """

        hour_types_response = self.get(
            url=f"https://{self._domain}/api/svc/core/diariesquery/diarysumaries/workday/diaryhourtypes",
            params={
                "userId": self._user_id,
                "date": date
            }
        )

        if hour_types_response.status == 200:
            return hour_types_response.json()['diaryHourTypes']
    
        else:
            logger.error(f"Can't retrieve hour types for date {date}!")
            return {}
    

    def _get_workday_slots(self, diary_summary_id: int) -> dict:
        """
        Return the workday slots for a given day. Each slot is comprised by the following keys: "in, "out" and "motive"
        params:
        diary_summary_id: int. It can be retrieved via `get_presence`; each diary entry has its own `diarySummaryId` key.
        """

        workday_slots_response = self.get(
            url=f"https://{self._domain}/api/svc/core/diariesquery/diarysummaries/{diary_summary_id}/workday/slots/self"
        )

        if workday_slots_response.status == 200:
            return workday_slots_response.json()['slots']
    
        else:
            logger.error(f"Can't retrieve workday slots for diary entry {diary_summary_id}!")
            return {}


    def get_sign_requests(self, date: str):
        """
        Return the user requests for a given date, such as Holidays
        params:
        date: str. Sign requests date. WARNING! Date format must be "mm/dd/YYYY", this is different from the rest of queries.
        """
        # TODO: unify date argument to the same format as the rest of methods and reformat it before sending the GET request

        sign_motives_response = self.get(
            url=f"https://{self._domain}/api/svc/core/diary/user/requests",
            params={
                "date": date
            }
        )

        if sign_motives_response.status == 200:
            return sign_motives_response.json()
    
        else:
            logger.error(f"Can't retrieve sign motives for date {date}!")
            return {}
        

    def get_status(self, only_running_clock: bool = False) -> tuple[timedelta, bool]:
        """
        Return the total amount of worked hours as well as whether we're signed-in or signed-out at the moment
        """
        signs_in_day = self.get(
            url=f"{self._woffu_api_url}/api/signs"
        ).json()

        # Initialize a timer and the running clock boolean
        total_time = timedelta()
        running_clock = False

        # Just return the las sign status
        if only_running_clock:
            return total_time, signs_in_day[-1]['SignIn'] if signs_in_day else running_clock

        # Go through all the signs.
        # Prepare current time with local timezone to use in case UTC offsets are wrong in Woffu
        # - We do it this way to take into account Daylight Saving Timezones (CET +1, CEST +2, for example)
        current_time = datetime.now(tz=self._localzone)

        for sign in signs_in_day:
            running_clock = sign['SignIn']
            sign_date = sign['TrueDate']
            # Split UtcTime and keep only the offset
            utc_offset = f"{sign['UtcTime'].split(' ')[1].zfill(3)}00"
            # WORKAROUND: Woffu stores incorrect UTC information when signing in from the Webapp, showing same local date on 'TrueDate' and 'UtcTime' but with no UTC offset (+0)
            # - In the time being, we'll be using our local timezone for any sign that comes with +0 in the UtcTime as we suspect Woffu DB is running on a server with the same local timezone as us
            if utc_offset == "+0000":
                utc_offset = current_time.strftime("%z")

            sign_date_timezoned = f"{sign_date}{utc_offset}"

            if running_clock:
                t1 = datetime.strptime(sign_date_timezoned, f"{DEFAULT_DATE_FORMAT}T%H:%M:%S.%f%z")
            else:
                t2 = datetime.strptime(sign_date_timezoned, f"{DEFAULT_DATE_FORMAT}T%H:%M:%S.%f%z")
                # Only update total_time when there's a sign-out
                total_time += (t2 - t1)
                logger.debug(f"Total time on closed signs: {total_time.total_seconds() / 3600}")
        
        logger.debug(f"Total time on closed signs: {total_time.total_seconds() / 3600}")

        # If clock is still running, add the remaining time
        if running_clock:
            logger.debug(f"Current time: {current_time.isoformat()}")
            total_time += (current_time - t1)

        # Log worked hours:
        hours, rem = divmod(total_time.total_seconds(), 3600)
        minutes, seconds = divmod(rem, 60)
        logger.info('Hours worked today: {:02d}:{:02d}:{:02d}'.format(int(hours), int(minutes), int(seconds)))
        logger.info(f"You're currently signed {'in' if running_clock else 'out'}.")
        return total_time, running_clock

    def sign(self, type: str = "", motive: str = ""):
        """
        Sign in/out on Woffu
        params:
        type: str. Can be "in", "out" or "empty". If empty, it will sign in/out without checking the current status.
        """
        # TODO: Include the sign-in reason in the sign-in ("Working from home", "Office", etc...)
        # Get current sign status
        _, signed_in = self.get_status(only_running_clock=True)

        # Evaluate current sign value against desired
        match type:
            case "in" | "out":
                requested_sign: bool = True if type == "in" else False
                if signed_in == requested_sign:
                    logger.warning(f"User is already signed {type}, skipping new sign.")
                    return
            case _:
                pass

        # Send sign request
        logger.info("Sending sign request...")
        # Get the current datetime with local timezone
        current_time = datetime.now(tz=self._localzone)

        # Get the actual time offset in minutes
        utc_offset = current_time.utcoffset()
        if utc_offset:
            timezone_offset =- int(utc_offset.total_seconds() / 60)
        else:
            timezone_offset = 0
        
        return self.post(
            url=f"https://{self._domain}/api/svc/signs/signs",
            # You don't need to send any data, just an empty JSON object.
            # The only known keys that are sent from the webpage are:
            # - agreementEventId: probably auto-generated on sign-in event, always empty on sign-out
            # - deviceId: like "WebApp"
            # - latitude: always Null
            # - longitude: always Null
            # - requestId: always Null
            # - timezoneOffset: matches the UTC offset of the timezone in minutes, and in inverse sign (-120 minutes for CEST, for example)
            json={
                #'StartDate': current_time.isoformat(sep='T', timespec='seconds'),
                #'EndDate': current_time.isoformat(sep='T', timespec='seconds'),
                #'timezoneOffset': timezone_offset,
                #'UserId': self._user_id
            }
        )
    

    def get_diary_hour_types_summary(self, from_date: str = "", to_date:str = "") -> dict:
        """
        Return a summary of the diary hour types other than the working ones, such as 'Extr. a compensar'
        """

        hour_types_summary: dict = {}

        from_dt = datetime.strptime(from_date, DEFAULT_DATE_FORMAT).astimezone(self._localzone)
        to_dt = datetime.strptime(to_date, DEFAULT_DATE_FORMAT).astimezone(self._localzone)

        for day in range(0, (to_dt - from_dt).days +1):
            date = from_dt + timedelta(days=day)
            date_str = date.strftime(DEFAULT_DATE_FORMAT)
            hour_types = self._get_diary_hour_types(date=date.strftime(date_str))

            hour_types_dict = {}

            for hour_type in hour_types:
                if hour_type['name'] not in hour_types_dict:
                    hour_types_dict[hour_type['name']] = hour_type['hours']
                else:
                    hour_types_dict[hour_type['name']] += hour_type['hours']

            hour_types_summary[date_str] = hour_types_dict

        return hour_types_summary
    

    def get_summary_report(self, from_date:str = "", to_date:str = "") -> dict:
        """
        Generate a summary report based on the information provided by the Presence endpoint.
        """
        diaries = self._get_presence(from_date=from_date, to_date=to_date)
        summary_report = {}

        current_time = datetime.now(tz=self._localzone)

        logger.info(f"Retrieving workday slots and extra hours...")
        for diary in diaries:
            summary_report[diary['date']] = {}
            event_report = summary_report[diary['date']]

            diary_summary_id = diary['diarySummaryId']

            # Retrieve diary slots
            slots = self._get_workday_slots(diary_summary_id=diary_summary_id)

            total_time: float = 0.0

            for slot in slots:
                if slot and 'motive' in slot and slot['motive']:
                    total_time += slot['motive']['hours']
                else:
                    logger.info(f"Slot of date {diary['date']} doesn't have motive. Using `in` and `out` keys instead...")
                    # Compute the difference between the sign in and sign out times
                    in_date = slot['in']['trueDate']
                    in_utc_offset = f"{slot['in']['utcTime'].split(' ')[1].zfill(3)}00"
                    if in_utc_offset == "+0000":
                        in_utc_offset = current_time.strftime("%z")

                    out_date = slot['out']['trueDate']
                    out_utc_offset = f"{slot['out']['utcTime'].split(' ')[1].zfill(3)}00"
                    if out_utc_offset == "+0000":
                        out_utc_offset = current_time.strftime("%z")
                                           
                    in_date_timezoned = f"{in_date}{in_utc_offset}"
                    out_date_timezoned = f"{out_date}{out_utc_offset}"

                    in_dt = datetime.strptime(in_date_timezoned, f"{DEFAULT_DATE_FORMAT}T%H:%M:%S%z")
                    out_dt = datetime.strptime(out_date_timezoned, f"{DEFAULT_DATE_FORMAT}T%H:%M:%S%z")
            
                    total_time += (out_dt - in_dt).total_seconds() / 3600

            event_report['work_hours'] = total_time

            # INVALID APPROACH: Woffu Presence page is inaccurate, we must work with the daily slots
            # # Retrieve the worked hours
            # if 'presenceEvents' in diary and diary['presenceEvents']:
            #     for event in diary['presenceEvents']:    
            #         if 'work_hours' not in event_report:
            #             event_report['work_hours'] = event['totalTime']
            #             logger.debug(f"Work hours initialized for date {diary['date']}")
            #         else:
            #             event_report['work_hours'] += event['totalTime']
            #             logger.debug(f"Work hours increased for date {diary['date']}")
            
            # Retrieve the other hour types
            if 'diaryHourTypes' in diary and diary['diaryHourTypes']:
                for hour_type in diary['diaryHourTypes']:
                    hour_type_id = hour_type['hourTypeId']
                    hour_type_name = self.hour_types_dict[hour_type_id] if hour_type_id in self.hour_types_dict else hour_type_id
                    if hour_type_name not in event_report:
                        event_report[hour_type_name] = hour_type['hours']
                    else:
                        event_report[hour_type_name] += hour_type['hours']

        return summary_report
    
    
    def export_summary_to_csv(self, summary_report:dict, from_date: str = "", to_date: str = "", output_path:Path = Path(Path.home() / "Documents/woffu/summary_reports"), delimiter:str=","):
        """
        Export the summary report to a CSV file
        """
        # Convert JSON to a proper array
        reports_list = []
        reports_header = set()

        search_date_range = False
        min_date = datetime.now(tz=self._localzone)
        max_date = datetime.strptime("2000-01-01",DEFAULT_DATE_FORMAT).astimezone(tz=self._localzone)

        if not from_date or not to_date:
            search_date_range = True
        
        else:
            min_date = datetime.strptime(from_date, DEFAULT_DATE_FORMAT).astimezone(tz=self._localzone)
            max_date = datetime.strptime(to_date, DEFAULT_DATE_FORMAT).astimezone(tz=self._localzone)

        for date, summary in summary_report.items():
            summary['date'] = date
            reports_list.append(summary)
            # Update the header in case there are new keys in the summaries
            reports_header.update(summary.keys())

            if search_date_range:
                date_dt = datetime.strptime(date, DEFAULT_DATE_FORMAT).astimezone(tz=self._localzone)
                if date_dt < min_date:
                    min_date = date_dt
                if date_dt > max_date:
                    max_date = date_dt

        logger.debug(f"CSV header = {reports_header}")
        logger.debug(f"Reports list: {reports_list}")

        # Prepare the CSV filename
        csv_filename = output_path / f"woffu_summary_report_from_{min_date.strftime(DEFAULT_DATE_FORMAT)}_to_{max_date.strftime(DEFAULT_DATE_FORMAT)}.csv"
        csv_filename.parent.mkdir(parents=True, exist_ok=True)

        with open(csv_filename, 'w', newline=os.linesep) as csvfile:
            # Prepare the CSV writer
            writer = csv.DictWriter(csvfile, fieldnames=sorted(reports_header), delimiter=delimiter, quoting=csv.QUOTE_STRINGS)
            writer.writeheader()

            # Write rows
            for report in reports_list:
                writer.writerow(report)

        logger.info(f"✅ CSV exported to {csv_filename}")



import json
from typing import Optional, Union, Dict, Iterable, Protocol, runtime_checkable, Any, cast, List
from unittest import TestCase
from unittest.mock import Mock, patch
from io import BytesIO
from urllib.error import HTTPError, URLError
from http.client import HTTPMessage
from urllib.request import Request, OpenerDirector
from src.woffu_client.stdrequests_session import Session, HTTPResponse
import pytest
import asyncio
import urllib.request

class DummyRawResponse:
    def __init__(self, status: int = 200, content: bytes = b"", headers: Optional[Dict[str, str]] = None) -> None:
        self._status: int = status
        self._content: bytes = content
        self._pos: int = 0
        self._headers: Dict[str, str] = headers or {}

    def read(self, size: int = -1) -> bytes:
        if size is None or size == -1:
            size = len(self._content) - self._pos
        data = self._content[self._pos : self._pos + size]
        self._pos += size
        return data

    def getcode(self) -> int:
        return self._status

    def getheaders(self) -> list[tuple[str, str]]:
        return list(self._headers.items())

    def close(self) -> None:
        pass


class DummyOpener(OpenerDirector):
    _response: Optional[object]
    called_with: Optional[Request]

    def __init__(self) -> None:
        super().__init__()
        self.called_with: Optional[urllib.request.Request] = None
        self._response = None

    def open(self, req: Request, timeout: Optional[Union[int, float]] = None) -> object:
        self.called_with = req
        if self._response is None:
            raise RuntimeError("No response set for DummyOpener")
        return self._response


class DummyHTTPError(HTTPError):
    def __init__(self, url: str, code: int, msg: str, hdrs: Optional[HTTPMessage], fp: Optional[BytesIO] = None) -> None:
        if fp is None:
            fp = BytesIO(b"")
        # If hdrs is None, pass an empty HTTPMessage to avoid Pylance error
        hdrs_non_none = hdrs if hdrs is not None else HTTPMessage()
        super().__init__(url, code, msg, hdrs_non_none, fp)


class MalformedStr(str):
    def split(self, sep=None, maxsplit=-1):
        if sep == "charset=":
            raise Exception("forced exception")
        return super().split(sep, maxsplit)


# Define a Protocol for something that supports .read()
@runtime_checkable
class SupportsRead(Protocol):
    def read(self, n: int = -1) -> bytes: ...


class TestHTTPResponse(TestCase):
    def test_text_and_json(self) -> None:
        content = b'{"key": "value"}'
        headers = {"Content-Type": "application/json"}
        raw = DummyRawResponse(content=content, headers=headers)
        resp = HTTPResponse(raw, 200, headers)
        self.assertEqual(resp.text(), content.decode("utf-8"))
        self.assertEqual(resp.json(), {"key": "value"})

    def test_content_and_iter_content(self) -> None:
        content = b"abcdefg"
        raw = DummyRawResponse(content=content)
        resp = HTTPResponse(raw, 200, {}, stream=False)
        self.assertEqual(resp.content(), content)
        chunks = list(resp.iter_content(chunk_size=3))
        self.assertEqual(b"".join(chunks), content)

    def test_iter_content_streaming_and_none_chunk_size(self) -> None:
        content = b"abcdefgh"
        raw = DummyRawResponse(content=content)
        resp = HTTPResponse(raw, 200, {}, stream=True)
        chunks = list(resp.iter_content(chunk_size=None))
        self.assertEqual(b"".join(chunks), content)

    def test_json_invalid_raises(self) -> None:
        raw = DummyRawResponse(content=b"not json", headers={"Content-Type": "application/json"})
        resp = HTTPResponse(raw, 200, {})
        with self.assertRaises(json.JSONDecodeError):
            resp.json()

    def test_text_with_charset(self) -> None:
        content = "café".encode("latin-1")
        raw = DummyRawResponse(content=content, headers={"Content-Type": "text/plain; charset=latin-1"})
        resp = HTTPResponse(raw, 200, {"Content-Type": "text/plain; charset=latin-1"})
        self.assertEqual(resp.text(), "café")

    def test_text_with_no_content_type_header(self) -> None:
        content = "hello"
        raw = DummyRawResponse(content=content.encode("utf-8"))
        resp = HTTPResponse(raw, 200, {})
        self.assertEqual(resp.text(), content)

    def test_iter_content_zero_chunk_size(self) -> None:
        content = b"abcdef"
        raw = DummyRawResponse(content=content)
        resp = HTTPResponse(raw, 200, {}, stream=True)
        chunks = list(resp.iter_content(chunk_size=0))
        self.assertEqual(b"".join(chunks), b"")

    def test_headers_property_returns_dict(self) -> None:
        content = b"data"
        headers = {"X-Test": "value"}
        raw = DummyRawResponse(content=content, headers=headers)
        resp = HTTPResponse(raw, 200, headers)
        self.assertEqual(resp.headers, headers)

    def test_close_method_does_not_raise(self) -> None:
        content = b"test"
        raw = DummyRawResponse(content=content)
        resp = HTTPResponse(raw, 200, {})
        try:
            resp.close()
        except Exception as e:
            self.fail(f"HTTPResponse.close() raised an exception: {e}")

    def test_headers_case_insensitivity(self) -> None:
        content = b"data"
        headers = {"content-type": "application/json", "X-CUSTOM": "value"}
        raw = DummyRawResponse(content=content, headers=headers)
        resp = HTTPResponse(raw, 200, headers)
        # Should return keys as-is (dict), but check lookup is case-sensitive or not
        self.assertEqual(resp.headers.get("content-type"), "application/json")
        self.assertEqual(resp.headers.get("X-CUSTOM"), "value")
        # Also check fallback when key missing returns None
        self.assertIsNone(resp.headers.get("Non-Existent"))

    def test_iter_content_with_large_chunk_size(self) -> None:
        content = b"abc"
        raw = DummyRawResponse(content=content)
        resp = HTTPResponse(raw, 200, {}, stream=True)
        chunks = list(resp.iter_content(chunk_size=10))  # chunk_size > content length
        self.assertEqual(b"".join(chunks), content)

    def test_httpresponse_text_with_malformed_charset(self) -> None:
        headers = {"Content-Type": "text/html; charset="}  # empty charset part causes split[1] to be empty string
        resp = HTTPResponse(DummyRawResponse(content=b"abc"), status=200, headers=headers)
        # Should not raise, should fall back to default 'utf-8'
        assert resp.text() == "abc"

    def test_iter_content_chunk_size_variants(self) -> None:
        # _raw is None
        resp2 = HTTPResponse(None, 200, {}, stream=True)
        chunks2 = list(resp2.iter_content())
        assert chunks2 == [resp2.content()]

        # chunk_size <=0 yields b""
        resp3 = HTTPResponse(DummyRawResponse(), 200, {}, stream=True)
        chunks3 = list(resp3.iter_content(chunk_size=0))
        assert chunks3 == [b""]

    def test_httpresponse_text_malformed_charset(self) -> None:
        headers = cast(dict[str, str], {"Content-Type": MalformedStr("text/html; charset=utf-8")})
        resp = HTTPResponse(raw_resp=BytesIO(b"test"), status=200, headers=headers)
        # It should not raise, fallback encoding 'utf-8' used
        text = resp.text()
        assert text == "test"

    def test_aiter_content_chunk_size_fallback(self) -> None:
        class DummyRaw:
            def __init__(self):
                self.read_calls: List[int] = []

            def read(self, size: int) -> bytes:
                self.read_calls.append(size)
                # Return empty bytes immediately to stop iteration
                return b""

        raw = DummyRaw()
        response = HTTPResponse(raw_resp=raw, status=200, headers={}, stream=True)

        async def run_test() -> None:
            # chunk_size = None (should fall back to 8192)
            chunks = [chunk async for chunk in response.aiter_content(chunk_size=None)]
            self.assertEqual(chunks, [])
            self.assertEqual(raw.read_calls, [8192])

            # Reset calls
            raw.read_calls.clear()

            # chunk_size <= 0 (e.g., 0) (should fallback to 8192)
            chunks = [chunk async for chunk in response.aiter_content(chunk_size=0)]
            self.assertEqual(chunks, [])
            self.assertEqual(raw.read_calls, [8192])

        asyncio.run(run_test())

class TestSession(TestCase):

    # Helper function (put near top of file or inside TestSession)
    @staticmethod
    def get_request_data_bytes(req: Request) -> bytes:
        if req.data is None:
            return b""
        if isinstance(req.data, bytes):
            return req.data
        if isinstance(req.data, str):
            return req.data.encode("utf-8")
        if hasattr(req.data, "read"):
            file_like = cast(SupportsRead, req.data)
            result = file_like.read()
            if not isinstance(result, bytes):
                raise TypeError(f"Expected bytes from read(), got {type(result)}")
            return result
        if isinstance(req.data, Iterable):
            return b"".join(req.data)
        raise TypeError(f"Unsupported type for req.data: {type(req.data)}")

    def setUp(self) -> None:
        self.session: Session = Session()
        self.opener: DummyOpener = DummyOpener()
        self.session.opener = self.opener
        self.called_method: str = ""
        self.called_url: str = ""

    def _run_async(self, coro: Any) -> Any:
        """Helper to run async coroutines in sync test methods."""
        return asyncio.run(coro)
    
    def _patch_async_request(self, return_status: int, return_text: str) -> None:
        """Helper to replace async_request on self.session."""

        async def fake_async_request(method: str, url: str, **kwargs: Any) -> Any:
            self.called_method = method
            self.called_url = url

            class FakeResp:
                status: int = return_status

                def text(self) -> str:
                    return return_text

            return FakeResp()

        # Patch the method directly on the session instance
        self.session.async_request = fake_async_request  # type: ignore

    def test_request_success(self) -> None:
        dummy_content = b"Hello, world!"
        dummy_headers = {"Content-Type": "text/plain"}
        dummy_response = DummyRawResponse(status=200, content=dummy_content, headers=dummy_headers)
        self.opener._response = dummy_response

        resp = self.session.get("http://example.com")
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.content(), dummy_content)
        self.assertEqual(resp.headers, dummy_headers)
        assert self.opener.called_with is not None
        self.assertEqual(self.opener.called_with.get_full_url(), "http://example.com")

    def test_http_error_response(self) -> None:
        dummy_headers = HTTPMessage()
        dummy_headers.add_header("Content-Type", "text/plain")
        dummy_fp = BytesIO(b"error content")
        dummy_error = DummyHTTPError(
            url="http://example.com/error",
            code=404,
            msg="Not Found",
            hdrs=dummy_headers,
            fp=dummy_fp,
        )

        def raise_http_error(req: Request, timeout: Optional[Union[int, float]] = None) -> None:
            raise dummy_error

        self.opener.open = raise_http_error
        resp = self.session.get("http://example.com/error")
        self.assertEqual(resp.status, 404)
        self.assertEqual(resp.content(), b"error content")
        self.assertEqual(resp.headers.get("Content-Type"), "text/plain")

    def test_http_error_without_fp(self) -> None:
        dummy_headers = HTTPMessage()
        dummy_headers.add_header("X-Test", "yes")
        dummy_error = DummyHTTPError(
            url="http://example.com/error",
            code=500,
            msg="Server Error",
            hdrs=dummy_headers,
            fp=None
        )

        def raise_http_error(req: Request, timeout: Optional[Union[int, float]] = None) -> None:
            raise dummy_error

        self.opener.open = raise_http_error
        resp = self.session.get("http://example.com/error")
        self.assertEqual(resp.status, 500)
        self.assertEqual(resp.headers.get("X-Test"), "yes")

    def test_other_exception_handling(self) -> None:
        def raise_generic(req: Request, timeout: Optional[Union[int, float]] = None) -> None:
            raise ValueError("Unexpected")

        self.opener.open = raise_generic
        with self.assertRaises(ValueError):
            self.session.get("http://example.com")

    def test_post_put_delete_methods(self) -> None:
        self.opener._response = DummyRawResponse(status=201, content=b"ok")
        self.assertEqual(self.session.post("http://x.com", data=b"abc").status, 201)
        self.assertEqual(self.session.put("http://x.com", data=b"abc").status, 201)
        self.assertEqual(self.session.delete("http://x.com").status, 201)

    def test_streaming_request(self) -> None:
        content = b"streamed data"
        self.opener._response = DummyRawResponse(status=200, content=content)
        resp = self.session.get("http://example.com", stream=True)
        self.assertEqual(b"".join(resp.iter_content(5)), content)

    def test_request_with_headers_and_timeout(self) -> None:
        self.opener._response = DummyRawResponse(status=200, content=b"ok")
        resp = self.session.get("http://example.com", headers={"X-Test": "yes"}, timeout=5)
        self.assertEqual(resp.status, 200)

    def test_request_with_no_data_and_headers(self) -> None:
        self.opener._response = DummyRawResponse(status=200, content=b"done")
        resp = self.session.post("http://example.com", data=None, headers=None)
        self.assertEqual(resp.status, 200)

    def test_request_timeout_argument(self) -> None:
        dummy_content = b"timeout test"
        dummy_response = DummyRawResponse(status=200, content=dummy_content)
        self.opener._response = dummy_response
        resp = self.session.get("http://timeout.com", timeout=10)
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.content(), dummy_content)

    def test_request_with_unusual_headers(self) -> None:
        headers = {"X-Custom": "abc", "Content-Length": "10"}
        self.opener._response = DummyRawResponse(status=200, content=b"x" * 10, headers=headers)
        resp = self.session.get("http://example.com", headers=headers)
        self.assertEqual(resp.headers.get("X-Custom"), "abc")
        self.assertEqual(resp.headers.get("Content-Length"), "10")

    def test_request_streaming_false_and_true(self) -> None:
        self.opener._response = DummyRawResponse(status=200, content=b"streamtest")
        resp = self.session.get("http://example.com", stream=False)
        self.assertEqual(resp.content(), b"streamtest")

        self.opener._response = DummyRawResponse(status=200, content=b"streamtest")
        resp = self.session.get("http://example.com", stream=True)
        self.assertEqual(b"".join(resp.iter_content(5)), b"streamtest")

    def test_request_with_string_data(self) -> None:
        self.opener._response = DummyRawResponse(status=200, content=b"ok")
        # Assuming Session supports string data and encodes it internally
        resp = self.session.post("http://example.com", data="string data")
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.content(), b"ok")

        # Check that opener was called with a Request object containing the data
        called_req = self.opener.called_with
        self.assertIsNotNone(called_req)

        data_bytes: bytes = b""
        if called_req is not None:
            raw_data: Any = called_req.data
            if raw_data is None:
                data_bytes = b""
            elif isinstance(raw_data, bytes):
                data_bytes = raw_data
            elif isinstance(raw_data, SupportsRead):
                data_bytes = raw_data.read()
            elif isinstance(raw_data, Iterable):
                data_bytes = b"".join(raw_data)
            else:
                try:
                    data_bytes = bytes(raw_data)
                except Exception:
                    data_bytes = b""

        self.assertIn(b"string data", data_bytes)

    def test_request_with_file_like_data(self) -> None:
        # file-like object with read() returning bytes
        file_like = BytesIO(b"file contents")
        self.opener._response = DummyRawResponse(status=200, content=b"ok")
        resp = self.session.post("http://example.com", data=file_like)
        assert resp.status == 200  # or your dummy success

    def test_request_with_file_like_read_returns_non_bytes(self) -> None:
        class BadFileLike:
            def read(self):
                return "not bytes"
        with pytest.raises(TypeError):
            self.session.post("http://example.com", data=BadFileLike())

    def test_request_with_empty_headers_dict(self) -> None:
        self.opener._response = DummyRawResponse(status=200, content=b"ok")
        resp = self.session.get("http://example.com", headers={})
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.content(), b"ok")

    def test_request_with_multiple_headers(self) -> None:
        # Python dict can't have duplicate keys; so simulate with multiple calls or with special header format
        # We'll test with comma-separated header value as HTTP supports that
        headers = {"X-Test": "value1, value2"}
        self.opener._response = DummyRawResponse(status=200, content=b"ok", headers=headers)
        resp = self.session.get("http://example.com", headers=headers)
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.headers.get("X-Test"), "value1, value2")

    def test_request_with_custom_user_agent_header(self) -> None:
        headers = {"User-Agent": "MyTestAgent/1.0"}
        self.opener._response = DummyRawResponse(status=200, content=b"ok", headers=headers)
        resp = self.session.get("http://example.com", headers=headers)
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.headers.get("User-Agent"), "MyTestAgent/1.0")

    def test_session_close_method_exists(self) -> None:
        # Just call close on session to check no error (assuming Session has close)
        try:
            self.session.close()
        except Exception as e:
            self.fail(f"Session.close() raised an exception: {e}")

    def test_apply_auth_header_adds_header(self) -> None:
        headers = {}
        self.session._apply_auth_header(headers, ("user", "pass"))
        assert "Authorization" in headers
        assert headers["Authorization"].startswith("Basic ")

    def test_request_data_types_and_invalid(self) -> None:
        with patch.object(self.session.opener, "open", return_value=Mock(getcode=lambda:200, getheaders=lambda:[], read=lambda: b"ok")):
            # dict data
            resp = self.session.request("POST", "http://example.com", data={"key": "val"})
            assert resp.status == 200
            # str data
            resp = self.session.request("POST", "http://example.com", data="stringdata")
            assert resp.status == 200
            # bytes data
            resp = self.session.request("POST", "http://example.com", data=b"bytesdata")
            assert resp.status == 200
            # invalid data
            with pytest.raises(TypeError):
                self.session.request("POST", "http://example.com", data=12345) # type: ignore[arg-type]
    
    def test_request_http_error_handling(self) -> None:
        
        headers = HTTPMessage()
        headers.add_header("X-Test", "1")

        error = HTTPError("http://example.com", 404, "Not Found", hdrs=headers, fp=None)
        with patch.object(self.session.opener, "open", side_effect=error):
            resp = self.session.request("GET", "http://example.com")
            assert resp.status == 404
            assert resp.headers.get("X-Test") == "1"


    def test_request_retries_and_raises(self) -> None:
        s = Session(retries=2)
        err = URLError("fail")
        call_count = 0

        def side_effect(*args, **kwargs):
            nonlocal call_count
            call_count += 1
            raise err

        with patch.object(s.opener, "open", side_effect=side_effect):
            with pytest.raises(URLError):
                s.request("GET", "http://example.com")
        assert call_count == 2

    def test_async_request_methods(self):
        resp_mock = Mock(status=200)
        
        async def run_test():
            with patch.object(self.session, "request", lambda *a, **kw: resp_mock):
                r = await self.session.async_get("http://example.com")
                self.assertEqual(r, resp_mock)
        
        asyncio.run(run_test())

    def test_context_manager_and_close_methods(self) -> None:
        # test __enter__ and __exit__
        with self.session as sess:
            assert sess is self.session

        # test close clears cookies and closes opener
        self.session._cookie_jar.set_cookie(Mock())
        close_mock = Mock()
        self.session.opener.close = close_mock
        self.session.close()
        assert close_mock.called

    def test_request_with_iterable_data(self) -> None:
        data = (chunk for chunk in [b"chunk1", b"chunk2"])
        self.opener._response = DummyRawResponse(status=200, content=b"ok")
        resp = self.session.post("http://example.com", data=data)
        self.assertEqual(resp.status, 200)
        called_req = self.opener.called_with
        self.assertIsNotNone(called_req)
        # The data should be bytes of concatenated chunks
        if called_req is not None:
            self.assertIn(b"chunk1chunk2", called_req.data if isinstance(called_req.data, bytes) else b"")

    def test_request_with_auth_argument(self) -> None:
        self.opener._response = DummyRawResponse(status=200, content=b"ok")
        resp = self.session.get("http://example.com", auth=("user", "pass"))
        self.assertEqual(resp.status, 200)
        called_req = self.opener.called_with
        self.assertIsNotNone(called_req)
        # Authorization header should be present in the request
        if called_req is not None:
            self.assertTrue("Authorization" in called_req.headers or "authorization" in called_req.headers)

    def test_request_passes_timeout(self) -> None:
        dummy_response = DummyRawResponse(status=200, content=b"ok")
        self.opener._response = dummy_response
        resp = self.session.request("GET", "http://example.com", timeout=7)
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.content(), b"ok")

    def test_request_with_headers_none(self) -> None:
        self.opener._response = DummyRawResponse(status=200, content=b"ok")
        resp = self.session.request("GET", "http://example.com", headers=None)
        self.assertEqual(resp.status, 200)
        self.assertEqual(resp.content(), b"ok")

    def test_request_with_params(self) -> None:
        called_urls: List[str] = []

        # Match the signature of OpenerDirector.open(fullurl, data=None, timeout=None)
        def fake_open(fullurl, data=None, timeout=None):
            # fullurl can be str or Request object
            url = fullurl.full_url if hasattr(fullurl, "full_url") else fullurl
            called_urls.append(url)

            class FakeRawResp:
                def getcode(self): return 200
                def getheaders(self): return []
                def read(self): return b"{}"

            return FakeRawResp()
    
        # Save original method to restore later
        original_open = self.session.opener.open

        try:
            # Replace the opener.open method with our fake_open
            self.session.opener.open = fake_open

            # URL without query
            resp = self.session.request("GET", "http://example.com/api", params={"a": "1", "b": "2"})
            self.assertIn("?", called_urls[-1])
            self.assertIn("a=1", called_urls[-1])
            self.assertIn("b=2", called_urls[-1])

            # URL with existing query
            resp = self.session.request("GET", "http://example.com/api?x=5", params={"a": "1"})
            self.assertIn("&", called_urls[-1])
            self.assertIn("x=5", called_urls[-1])
        finally:
            self.session.opener.open = original_open

    def test_request_defensive_fallback_range_empty(self):

        # Define a local 'range' that yields nothing to simulate empty retry loop
        def empty_range(*args, **kwargs):
            return iter(())

        # Save original range for later restoration
        import builtins
        original_range = builtins.range

        try:
            # Override builtins.range locally inside this test
            builtins.range = empty_range

            with self.assertRaises(RuntimeError):
                self.session.request("GET", "http://example.com")
        finally:
            # Restore original range
            builtins.range = original_range

    def test_http_error_with_other_codes(self) -> None:
        for code in [400, 401, 500, 503]:
            headers = HTTPMessage()
            headers.add_header("X-Test", f"code-{code}")
            error = HTTPError("http://example.com", code, "Error", hdrs=headers, fp=None)
            with patch.object(self.session.opener, "open", side_effect=error):
                resp = self.session.request("GET", "http://example.com")
                self.assertEqual(resp.status, code)
                self.assertEqual(resp.headers.get("X-Test"), f"code-{code}")

    def test_async_get(self) -> None:
        self._patch_async_request(200, "get_ok")

        async def run_test() -> None:
            resp = await self.session.async_get("http://example.com")
            self.assertEqual(resp.status, 200)
            self.assertEqual(resp.text(), "get_ok")
            self.assertEqual(self.called_method, "GET")
            self.assertEqual(self.called_url, "http://example.com")

        self._run_async(run_test())

    def test_async_post(self) -> None:
        self._patch_async_request(201, "post_ok")

        async def run_test() -> None:
            resp = await self.session.async_post("http://example.com", data={"key": "value"})
            self.assertEqual(resp.status, 201)
            self.assertEqual(resp.text(), "post_ok")
            self.assertEqual(self.called_method, "POST")
            self.assertEqual(self.called_url, "http://example.com")

        self._run_async(run_test())

    def test_async_put(self) -> None:
        self._patch_async_request(202, "put_ok")

        async def run_test() -> None:
            resp = await self.session.async_put("http://example.com")
            self.assertEqual(resp.status, 202)
            self.assertEqual(resp.text(), "put_ok")
            self.assertEqual(self.called_method, "PUT")
            self.assertEqual(self.called_url, "http://example.com")

        self._run_async(run_test())

    def test_async_patch(self) -> None:
        self._patch_async_request(200, "patch_ok")

        async def run_test() -> None:
            resp = await self.session.async_patch("http://example.com")
            self.assertEqual(resp.status, 200)
            self.assertEqual(resp.text(), "patch_ok")
            self.assertEqual(self.called_method, "PATCH")
            self.assertEqual(self.called_url, "http://example.com")

        self._run_async(run_test())

    def test_async_delete(self) -> None:
        self._patch_async_request(204, "delete_ok")

        async def run_test() -> None:
            resp = await self.session.async_delete("http://example.com")
            self.assertEqual(resp.status, 204)
            self.assertEqual(resp.text(), "delete_ok")
            self.assertEqual(self.called_method, "DELETE")
            self.assertEqual(self.called_url, "http://example.com")

        self._run_async(run_test())

    def test_patch_method_calls_request(self) -> None:
        # Prepare a dummy raw response to use in DummyOpener
        self.opener._response = DummyRawResponse()  # set dummy response so open() won't raise

        with patch.object(self.session, "request", wraps=self.session.request) as mock_request:
            # Call patch
            resp = self.session.patch("http://example.com/api", data={"key": "value"})

            # Assert request was called once with method="PATCH"
            mock_request.assert_called_once()
            args, kwargs = mock_request.call_args
            self.assertEqual(args[0], "PATCH")  # method argument
            self.assertEqual(args[1], "http://example.com/api")  # url argument
            self.assertIn("data", kwargs)
            self.assertEqual(kwargs["data"], {"key": "value"})

            # Since the patched request returns an HTTPResponse, resp should be an HTTPResponse instance
            self.assertIsInstance(resp, HTTPResponse)

    def test_json_serialization_sets_header_and_body(self) -> None:
        expected = {"hello": "world"}
        self.opener._response = DummyRawResponse(
            status=200, content=b"{}", headers={"Content-Type": "application/json"}
            )

        resp: HTTPResponse = self.session.post("http://example.com", json=expected)
        self.assertIsInstance(resp, HTTPResponse)

        req: Request = cast(urllib.request.Request, self.opener.called_with)
        self.assertIsNotNone(req)

        # Careful! Python's urllib.request.Request internally stores the header keys with the capitalization from when they were added,
        # and `get_header()` is case-sensitive except for "Content-type", which it seems to be a weird historical quirk.
        #self.assertEqual(req.get_header("Content-type"), "custom/type")
        
        # For robustness's sake, it's better to normalize headers when asserting instead.
        headers = {k.lower(): v for k, v in req.header_items()}
        self.assertEqual(headers["content-type"], "application/json")

        # Safely get data bytes and decode
        data_bytes = self.get_request_data_bytes(req)
        self.assertEqual(data_bytes.decode(), json.dumps(expected))

    def test_json_respects_custom_header(self) -> None:
        expected = {"foo": "bar"}
        self.opener._response = DummyRawResponse(
            status=200, content=b"{}", headers={"Content-Type": "application/json"}
            )

        resp: HTTPResponse = self.session.post(
            "http://example.com",
            json=expected,
            headers={"Content-Type": "custom/type"},
        )
        self.assertIsInstance(resp, HTTPResponse)

        req: Request = cast(urllib.request.Request, self.opener.called_with)
        self.assertIsNotNone(req)

        # Careful! Python's urllib.request.Request internally stores the header keys with the capitalization from when they were added,
        # and `get_header()` is case-sensitive except for "Content-type", which it seems to be a weird historical quirk.
        #self.assertEqual(req.get_header("Content-type"), "custom/type")
        
        # For robustness's sake, it's better to normalize headers when asserting instead.
        headers = {k.lower(): v for k, v in req.header_items()}
        self.assertEqual(headers["content-type"], "custom/type")
        
        # Safely get data bytes and decode
        data_bytes = self.get_request_data_bytes(req)
        self.assertEqual(data_bytes.decode(), json.dumps(expected))

    def test_json_none_means_no_body(self) -> None:
        self.opener._response = DummyRawResponse(
            status=200, content=b"{}", headers={"Content-Type": "application/json"}
            )

        resp = self.session.post("http://example.com", json=None)
        self.assertIsInstance(resp, HTTPResponse)

        req: Request = cast(Request, self.opener.called_with)
        self.assertIsNotNone(req)

        # Should send no body at all
        self.assertIsNone(req.data)